import ujson

import datetime
import dateutil.parser
from inflection import camelize
from rest_framework import status
from rest_framework.reverse import reverse

from .tests import USER, STAFF


class ResourceCreateTestCase(object):

    def test_create__valid_data_incorrect_header(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data=self.json_request, content_type='application/json')

        self.failure_response_structure_test(response, status.HTTP_415_UNSUPPORTED_MEDIA_TYPE)

    def test_create__empty_request(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data={}, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_400_BAD_REQUEST)

    def test_create__missing_required_parameters(self):
        if not self.required_params:
            return True

        url = reverse(self.resource_view_name)
        self.create_json_format['data']['attributes'] = {'random_key': 'This is a random value'}

        missing_parameters_request = ujson.dumps(self.create_json_format)

        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data=missing_parameters_request, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_400_BAD_REQUEST)

    def test_create__without_relationships(self):
        if not self.relationship_keys:
            return True

        url = reverse(self.resource_view_name)
        self.create_json_format['data']['relationships'] = {}
        no_relationships_request = ujson.dumps(self.create_json_format)

        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data=no_relationships_request, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_400_BAD_REQUEST)

    def test_create__with_malformed_relationship_data(self):
        if not self.relationship_keys:
            return True

        url = reverse(self.resource_view_name)
        key = camelize(self.relationship_keys[0], False)
        self.create_json_format['data']['relationships'][key]['data'] = {
            # No type provided
            'id': '1234'}

        empty_relationships_request = ujson.dumps(self.create_json_format)

        response = self.client_post_auth(url, user_role=self.USER_ROLE, data=empty_relationships_request,
                                         content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_400_BAD_REQUEST)

    def test_create__post_non_json_data_correctly_errors(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data='invalid_JSON_obj', content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_400_BAD_REQUEST)

    def test_create__unauthorized(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, data=self.json_request, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)


class ResourceCreateWithoutPermissionTestCase(object):

    def test_create__valid_data_incorrect_header(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data=self.json_request, content_type='application/json')

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_create__empty_request(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data={}, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_create__missing_required_parameters(self):
        if not self.required_params:
            return True

        url = reverse(self.resource_view_name)
        self.create_json_format['data']['attributes'] = {'random_key': 'This is a random value'}

        missing_parameters_request = ujson.dumps(self.create_json_format)

        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data=missing_parameters_request, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_create__without_relationships(self):
        if not self.relationship_keys:
            return True

        url = reverse(self.resource_view_name)
        self.create_json_format['data']['relationships'] = {}
        no_relationships_request = ujson.dumps(self.create_json_format)

        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data=no_relationships_request, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_create__with_malformed_relationship_data(self):
        if not self.relationship_keys:
            return True

        url = reverse(self.resource_view_name)
        key = camelize(self.relationship_keys[0], False)
        self.create_json_format['data']['relationships'][key]['data'] = {
            # No type provided
            'id': '1234'}

        empty_relationships_request = ujson.dumps(self.create_json_format)

        response = self.client_post_auth(url, user_role=self.USER_ROLE, data=empty_relationships_request,
                                         content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_create__post_non_json_data_correctly_errors(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, user_role=self.USER_ROLE,
                                         data='invalid_JSON_obj', content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_create__unauthorized(self):
        url = reverse(self.resource_view_name)
        response = self.client_post_auth(url, data=self.json_request, content_type='application/vnd.api+json')

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)


class ResourceUpdateTestCase(object):

    def get_patch_response(self, request_data):
        user_id = getattr(getattr(self.resource, 'user', None), 'id', None)

        return self.client_patch_auth(
            reverse(self.resource_view_name, args=(self.resource.id,)),
            user_role=self.USER_ROLE,
            user_id=user_id,
            data=ujson.dumps(request_data),
            content_type='application/vnd.api+json'
        )

    def test_update__resource(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.get_patch_response(update_data)

        self.success_response_structure_test(response, status.HTTP_200_OK)

        setattr(self.resource, attribute_name, new_attribute_value)

        self.validate_instance_in_response(response, self.resource, self.attributes,
                                           relationship_keys=self.relationship_keys)

        db_obj = self.resource_class.objects.get(id=self.resource.id)
        db_value = getattr(db_obj, attribute_name)
        if isinstance(db_value, datetime.datetime):
            new_attribute_value = dateutil.parser.parse(new_attribute_value)

        elif isinstance(db_value, datetime.date):
            new_attribute_value = new_attribute_value.isoformat()

        elif isinstance(db_value, datetime.time):
            new_attribute_value = new_attribute_value.isoformat()

        self.assertEqual(db_value, new_attribute_value)

    def test_update__incorrect_type(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['type'] = 'RandomType'
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.get_patch_response(update_data)

        self.failure_response_structure_test(response, status.HTTP_409_CONFLICT)

    def test_update__multiple_fields(self):
        if len(self.new_attribute_values) < 2:
            return True

        attribute_name_1 = self.attributes[0]
        new_attribute_value_1 = self.new_attribute_values[0]

        attribute_name_2 = self.attributes[1]
        new_attribute_value_2 = self.new_attribute_values[1]

        update_data = self.patch_request_stub
        update_data['data']['attributes'] = {
            attribute_name_1: new_attribute_value_1,
            attribute_name_2: new_attribute_value_2
        }

        response = self.get_patch_response(update_data)

        setattr(self.resource, attribute_name_1, new_attribute_value_1)
        setattr(self.resource, attribute_name_2, new_attribute_value_2)

        self.success_response_structure_test(response, status.HTTP_200_OK)

        self.validate_instance_in_response(response, self.resource, self.attributes,
                                           relationship_keys=self.relationship_keys)

        db_obj = self.resource_class.objects.get(id=self.resource.id)
        self.assertEqual(getattr(db_obj, attribute_name_1), new_attribute_value_1)
        self.assertEqual(getattr(db_obj, attribute_name_2), new_attribute_value_2)

    def test_update__missing_item_404(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['id'] = 12457
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.client_patch_auth(
            reverse(self.resource_view_name, args=(12457,)),
            user_role=self.USER_ROLE,
            data=ujson.dumps(update_data),
            content_type='application/vnd.api+json'
        )

        self.failure_response_structure_test(response, status.HTTP_404_NOT_FOUND)

    def test_update__unauthorized(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.client_patch_auth(
            reverse(self.resource_view_name, args=(self.resource.id,)),
            data=ujson.dumps(update_data),
            content_type='application/vnd.api+json'
        )

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)


class ResourceUpdateLimitedPermissionTestCase(object):

    def get_patch_response(self, request_data):
        user_id = getattr(getattr(self.resource, 'user', None), 'id', None)

        return self.client_patch_auth(
            reverse(self.resource_view_name, args=(self.resource.id,)),
            user_role=self.USER_ROLE,
            user_id=user_id,
            data=ujson.dumps(request_data),
            content_type='application/vnd.api+json'
        )

    def test_update__resource(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.get_patch_response(update_data)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_update__incorrect_type(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['type'] = 'RandomType'
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.get_patch_response(update_data)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_update__multiple_fields(self):
        if len(self.new_attribute_values) < 2:
            return True

        attribute_name_1 = self.attributes[0]
        new_attribute_value_1 = self.new_attribute_values[0]

        attribute_name_2 = self.attributes[1]
        new_attribute_value_2 = self.new_attribute_values[1]

        update_data = self.patch_request_stub
        update_data['data']['attributes'] = {
            attribute_name_1: new_attribute_value_1,
            attribute_name_2: new_attribute_value_2
        }

        response = self.get_patch_response(update_data)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_update__missing_item_404(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['id'] = 12457
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.client_patch_auth(
            reverse(self.resource_view_name, args=(12457,)),
            user_role=self.USER_ROLE,
            data=ujson.dumps(update_data),
            content_type='application/vnd.api+json'
        )

        self.failure_response_structure_test(response, status.HTTP_404_NOT_FOUND)

    def test_update__unauthorized(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.client_patch_auth(
            reverse(self.resource_view_name, args=(self.resource.id,)),
            data=ujson.dumps(update_data),
            content_type='application/vnd.api+json'
        )

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)


class ResourceUpdateWithoutPermissionTestCase(ResourceUpdateLimitedPermissionTestCase):

    # Overriding method, when no permission we return a 403 instead of a 404
    def test_update__missing_item_404(self):
        attribute_name = self.attributes[0]
        new_attribute_value = self.new_attribute_values[0]

        update_data = self.patch_request_stub
        update_data['data']['id'] = 12457
        update_data['data']['attributes'] = {attribute_name: new_attribute_value}

        response = self.client_patch_auth(
            reverse(self.resource_view_name, args=(12457,)),
            user_role=self.USER_ROLE,
            data=ujson.dumps(update_data),
            content_type='application/vnd.api+json'
        )

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)


class ResourceFlaggedDeleteTestCase(object):

    def test_resource(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        # Verify that the object still exists and that the deleted_at field has been set
        deleted_resource = self.resource_class.objects.filter(id=self.resource.id)
        self.assertEqual(deleted_resource.count(), 1)
        self.assertIsNotNone(deleted_resource[0].deleted_at)

    def test_delete__404_for_nonexistant_resource(self):
        url = reverse(self.resource_view_name, args=(9999,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE)

        self.failure_response_structure_test(response, status.HTTP_404_NOT_FOUND)

    def test_delete__unauthorized(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url)

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)


class ResourceDeleteTestCase(object):

    def test_delete_resource__by_user(self):
        if not hasattr(self.resource, 'user'):
            return True
        resource = self.resource_class.objects.filter(id=self.resource.id)
        self.assertEqual(resource.count(), 1)

        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE, user_id=self.resource.user.id)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        deleted_resource = self.resource_class.objects.filter(id=self.resource.id)
        self.assertEqual(deleted_resource.count(), 0)

    def test_delete_resource__by_staff(self):
        resource = self.resource_class.objects.filter(id=self.resource.id)
        self.assertEqual(resource.count(), 1)

        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=STAFF)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        deleted_resource = self.resource_class.objects.filter(id=self.resource.id)
        self.assertEqual(deleted_resource.count(), 0)

    def test_delete__404_for_nonexistant_resource(self):
        url = reverse(self.resource_view_name, args=(9999,))
        response = self.client_delete_auth(url, user_role=STAFF)

        self.failure_response_structure_test(response, status.HTTP_404_NOT_FOUND)

    def test_delete__unauthorized(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url)

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)

    def test_delete__forbidden(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=USER)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)


class ResourceNoDeleteTestCase(object):

    def test_delete_resource__by_user(self):
        if not hasattr(self.resource, 'user'):
            return True
        resource = self.resource_class.objects.filter(id=self.resource.id)
        self.assertEqual(resource.count(), 1)

        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE, user_id=self.resource.user.id)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_delete_resource__by_staff(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_delete__for_nonexistant_resource(self):
        url = reverse(self.resource_view_name, args=(9999,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)

    def test_delete__unauthorized(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url)

        self.failure_response_structure_test(response, status.HTTP_401_UNAUTHORIZED)

    def test_delete__forbidden(self):
        url = reverse(self.resource_view_name, args=(self.resource.id,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE)

        self.failure_response_structure_test(response, status.HTTP_403_FORBIDDEN)


class ResourceDeleteWithoutPermissionTestCase(ResourceNoDeleteTestCase):
    pass


class ResourceDeleteLimitedTestCase(ResourceNoDeleteTestCase):

    def test_delete__for_nonexistant_resource(self):
        url = reverse(self.resource_view_name, args=(9999,))
        response = self.client_delete_auth(url, user_role=self.USER_ROLE)

        self.failure_response_structure_test(response, status.HTTP_404_NOT_FOUND)



from django.db.models import CharField, TextField
from django.db.models import Model
from django.db.models.manager import Manager
from django.db.models.query import QuerySet
from rest_framework import viewsets
from rest_framework.exceptions import MethodNotAllowed
from rest_framework_json_api.views import RelationshipView as OldRelView

from zc_common.remote_resource.models import RemoteResource
from zc_common.remote_resource.serializers import ResourceIdentifierObjectSerializer


class ModelViewSet(viewsets.ModelViewSet):
    """
    This class overwrites the ModelViewSet's list method, which handles
    requests made to the collection's base endpoint (/collection), in
    order to provide support for filtering via the filter[] query parameter.

    Inheriting from this class, along with adding the filter backend, will properly
    handle requests made to /collection as well as /collection?filter[name]=test.
    It's also possible to filter by a collection of primary keys, for example:
    /collection?filter[id__in]=1,2,3
    Requests to filter on keys that do not exist will return an empty set.
    """

    @property
    def filter_fields(self):
        queryset = self.get_queryset()
        return_fields = {}

        fields = queryset.model._meta.get_fields()
        for field in fields:
            # For backwards compatibility GenericForeignKey should not be
            # included in the results.
            if field.is_relation and field.many_to_one and field.related_model is None:
                continue
            # Relations to child proxy models should not be included.
            if (field.model != queryset.model._meta.model and
                    field.model._meta.concrete_model == queryset.model._meta.concrete_model):
                continue

            name = field.attname if hasattr(field, 'attname') else field.name
            if hasattr(field, 'primary_key') and field.primary_key:
                return_fields['id'] = ['in', 'exact']
            elif CharField in field.__class__.__mro__ or TextField in field.__class__.__mro__:
                return_fields[name] = ['icontains', 'exact']
            else:
                return_fields[name] = ['exact']

        return return_fields

    def has_ids_query_params(self):
        return hasattr(self.request, 'query_params') and 'filter[id__in]' in self.request.query_params


class RelationshipView(OldRelView):
    serializer_class = ResourceIdentifierObjectSerializer

    def patch(self, request, *args, **kwargs):
        """
        Restricting PATCH requests made to the relationship view temporarily to
        prevent the possibility of data corruption when PATCH requests are made
        to to-many related resources. This override will not be necessary
        once a fix is made upstream.

        See:
        https://github.com/django-json-api/django-rest-framework-json-api/issues/242
        """
        raise MethodNotAllowed('PATCH')

    def _instantiate_serializer(self, instance):
        if isinstance(instance, RemoteResource):
            return ResourceIdentifierObjectSerializer(instance=instance)

        if isinstance(instance, Model) or instance is None:
            return self.get_serializer(instance=instance)
        else:
            if isinstance(instance, (QuerySet, Manager)):
                instance = instance.all()

            return self.get_serializer(instance=instance, many=True)



import jwt
from django.utils import encoding
from rest_framework_jwt.settings import api_settings
from permissions import SERVICE_ROLES


def jwt_payload_handler(user):
    """Constructs a payload for a user JWT.
    This is a slimmed down version of the handler in
    https://github.com/GetBlimp/django-rest-framework-jwt/

    :param user: an object with `pk` and `get_roles()`
    :return: A dictionary that can be passed into `jwt_encode_handler`
    """

    payload = {
        'id': encoding.force_text(user.pk),
        'roles': user.get_roles(),
    }

    return payload


def service_jwt_payload_handler(service_name):
    """Constructs a payload for a service JWT.

    :param service_name: a string corresponding to the name of the service where this JWT will be sent
    :return: a dictionary that can be passed into `jwt_encode_handler`
    """
    payload = {
        'serviceName': service_name,
        'roles': SERVICE_ROLES
    }

    return payload


def jwt_encode_handler(payload):
    """Encodes a payload into a valid JWT.

    :param payload: a dictionary
    :return: an encoded JWT string
    """

    return jwt.encode(
        payload,
        api_settings.JWT_SECRET_KEY,
        api_settings.JWT_ALGORITHM
    ).decode('utf-8')



from datetime import date
import six
import time
import uuid

import boto
from boto.s3.key import Key
from django.conf import settings
from zc_common.events.emit import emit_microservice_event


S3_BUCKET_NAME = 'zc-mp-email'
EMAIL_EVENT_TYPE = 'send_email'
ATTACHMENT_PREFIX = 'attachment_'


class MissingCredentialsError(Exception):
    pass


def get_s3_email_bucket():
    aws_access_key_id = settings.AWS_ACCESS_KEY_ID
    aws_secret_access_key = settings.AWS_SECRET_ACCESS_KEY
    if not (aws_access_key_id and aws_secret_access_key):
        msg = 'You need to set AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY in your settings file.'
        raise MissingCredentialsError(msg)

    conn = boto.connect_s3(aws_access_key_id, aws_secret_access_key)
    bucket = conn.get_bucket(S3_BUCKET_NAME)
    return bucket


def generate_s3_folder_name(email_uuid):
    email_date = date.today().isoformat()
    email_timestamp = int(time.time())
    return "{}/{}_{}".format(email_date, email_timestamp, email_uuid)


def generate_s3_content_key(s3_folder_name, content_type, content_name=''):
    content_key = "{}/{}".format(s3_folder_name, content_type)
    if content_name:
        content_key += '_{}'.format(content_name)
    return content_key


def upload_string_to_s3(bucket, content_key, content):
    if content:
        k = Key(bucket)
        k.key = content_key
        k.set_contents_from_string(content)


def upload_file_to_s3(bucket, content_key, filename):
    if filename:
        k = Key(bucket)
        k.key = content_key
        k.set_contents_from_filename(filename)


def send_email(from_email=None, to=None, cc=None, bcc=None, reply_to=None,
               subject=None, plaintext_body=None, html_body=None, headers=None,
               files=None, attachments=None, user_id=None, resource_type=None, resource_id=None,
               logger=None):
    """
    files:       A list of file paths
    attachments: A list of tuples of the format (filename, content_type, content)
    """
    email_uuid = uuid.uuid4()
    bucket = get_s3_email_bucket()
    s3_folder_name = generate_s3_folder_name(email_uuid)
    if logger:
        msg = '''MICROSERVICE_SEND_EMAIL: Upload email with UUID {}, to {}, from {},
        with attachments {} and files {}'''
        logger.info(msg.format(email_uuid, to, from_email, attachments, files))

    to = to.split(',') if isinstance(to, six.string_types) else to
    cc = cc.split(',') if isinstance(cc, six.string_types) else cc
    bcc = bcc.split(',') if isinstance(bcc, six.string_types) else bcc
    reply_to = reply_to.split(',') if isinstance(reply_to, six.string_types) else reply_to
    for arg in (to, cc, bcc, reply_to):
        if arg and not isinstance(arg, list):
            msg = "Keyword arguments 'to', 'cc', 'bcc', and 'reply_to' should be of <type 'list'>"
            raise TypeError(msg)

    if not any([to, cc, bcc, reply_to]):
        msg = "Keyword arguments 'to', 'cc', 'bcc', and 'reply_to' can't all be empty"
        raise TypeError(msg)

    html_body_key = None
    if html_body:
        html_body_key = generate_s3_content_key(s3_folder_name, 'html')
        upload_string_to_s3(bucket, html_body_key, html_body)

    plaintext_body_key = None
    if plaintext_body:
        plaintext_body_key = generate_s3_content_key(s3_folder_name, 'plaintext')
        upload_string_to_s3(bucket, plaintext_body_key, plaintext_body)

    attachments_keys = []
    if attachments:
        for filename, mimetype, attachment in attachments:
            attachment_key = generate_s3_content_key(s3_folder_name, 'attachment',
                                                     content_name=filename)
            upload_string_to_s3(bucket, attachment_key, attachment)
            attachments_keys.append(attachment_key)
    if files:
        for filepath in files:
            filename = filepath.split('/')[-1]
            attachment_key = generate_s3_content_key(s3_folder_name, 'attachment',
                                                     content_name=filename)
            upload_file_to_s3(bucket, attachment_key, filepath)
            attachments_keys.append(attachment_key)

    event_data = {
        'from_email': from_email,
        'to': to,
        'cc': cc,
        'bcc': bcc,
        'reply_to': reply_to,
        'subject': subject,
        'plaintext_body_key': plaintext_body_key,
        'html_body_key': html_body_key,
        'attachments_keys': attachments_keys,
        'headers': headers,
        'user_id': user_id,
        'resource_type': resource_type,
        'resource_id': resource_id,
        'task_id': str(email_uuid)
    }

    if logger:
        logger.info('MICROSERVICE_SEND_EMAIL: Sent email with UUID {} and data {}'.format(
            email_uuid, event_data
        ))

    emit_microservice_event(EMAIL_EVENT_TYPE, **event_data)



from __future__ import absolute_import

import sys
from zc_common.remote_resource.request import make_service_request, get_route_from_fk, ServiceRequestException


class EmitEventFailure(Exception):
    pass


def emit_event(service_name, event_type, event_name, resource_type, resource_id, user_id, meta_data=None):
    try:
        create_event_url = get_route_from_fk(event_type)

        event_data = generate_event_data(
            event_type,
            event_name,
            resource_type,
            resource_id,
            user_id,
            meta_data=meta_data
        )

        response = make_service_request(service_name, create_event_url, 'post', event_data)
        return response
    except ServiceRequestException:
        message = ("Error emitting {} to {}. resource_type: {}, resource_id: {}, user_id: {}, meta_data: {}"
                   .format(event_name, event_type, resource_type, resource_id, user_id, meta_data))
        raise EmitEventFailure(message), None, sys.exc_info()[2]


def generate_event_data(event_type, event_name, resource_type, resource_id, user_id, meta_data=None):
    data = {
        'data': {
            'type': event_type,
            'attributes': {
                'eventType': event_name,
                'resourceType': resource_type,
                'resourceId': resource_id,
                'userId': user_id
            }
        }
    }

    if meta_data and isinstance(meta_data, dict):
        data.update({
            'meta': meta_data
        })

    return data



from django.conf import settings
from rest_framework.settings import APISettings
import os

DEFAULTS = {
    'GATEWAY_ROOT_PATH': getattr(
        settings, 'GATEWAY_ROOT_PATH', os.environ.get('GATEWAY_ROOT_PATH', 'http://gateway:4000/'))
}

zc_settings = APISettings(None, DEFAULTS, None)



import logging
import uuid

from django.conf import settings


logger = logging.getLogger('django')


class EmitEventException(Exception):
    pass


def emit_microservice_event(event_type, *args, **kwargs):
    import pika
    import ujson

    url = settings.BROKER_URL

    params = pika.URLParameters(url)
    params.socket_timeout = 5

    event_queue_name = '{}-events'.format(settings.SERVICE_NAME)

    connection = pika.BlockingConnection(params)

    channel = connection.channel()
    channel.queue_declare(queue=event_queue_name, durable=True)

    task_id = str(uuid.uuid4())

    keyword_args = {'task_id': task_id}
    keyword_args.update(kwargs)

    message = {
        'task': 'microservice.event',
        'id': task_id,
        'args': [event_type] + list(args),
        'kwargs': keyword_args
    }

    event_body = ujson.dumps(message)

    logger.info('MICROSERVICE_EVENT::EMIT: Emitting [{}:{}] event for object ({}:{}) and user {}'.format(
        event_type, task_id, kwargs.get('resource_type'), kwargs.get('resource_id'),
        kwargs.get('user_id')))

    response = channel.basic_publish('microservice-events', '', event_body, pika.BasicProperties(
        content_type='application/json', content_encoding='utf-8'))

    if not response:
        logger.info(
            'MICROSERVICE_EVENT::EMIT_FAILURE: Failure emitting [{}:{}] event for object ({}:{}) and user {}'.format(
                event_type, task_id, kwargs.get('resource_type'), kwargs.get('resource_id'), kwargs.get('user_id')))
        raise EmitEventException("Message may have failed to deliver")

    return response



import ujson
import urllib

import requests
from inflection import underscore
from uritemplate import expand

from zc_common.jwt_auth.utils import service_jwt_payload_handler, jwt_encode_handler
from zc_common.settings import zc_settings


# Requests that can be made to another service
GET = 'get'
POST = 'post'
PUT = 'put'
PATCH = 'patch'


class UnsupportedHTTPMethodException(Exception):
    pass


class RouteNotFoundException(Exception):
    pass


class ServiceRequestException(Exception):
    pass


class RemoteResourceException(Exception):
    pass


class RemoteResourceWrapper(object):

    def __init__(self, data, included=None):
        result = self._get_from_include(included, data)
        self.data = result if result else data
        self.create_properties_from_data(included)

    def _get_from_include(self, included, obj):
        if included:
            res = included.get((obj['type'], obj['id']))
            return res
        return None

    def create_properties_from_data(self, included):
        accepted_keys = ('id', 'type', 'self', 'related')

        for key in self.data.keys():
            if key in accepted_keys:
                setattr(self, key, self.data.get(key))

        if 'attributes' in self.data:
            attributes = self.data['attributes']
            for key in attributes.keys():
                setattr(self, underscore(key), attributes[key])

        if 'relationships' in self.data:
            relationships = self.data['relationships']

            for key in relationships.keys():
                if isinstance(relationships[key]['data'], list):
                    setattr(self, underscore(key), RemoteResourceListWrapper(relationships[key]['data'], included))
                else:
                    got = None
                    if included:
                        got = self._get_from_include(included, relationships[key]['data'])

                    if got:
                        setattr(self, underscore(key), RemoteResourceWrapper(got, included))
                    else:
                        setattr(self, underscore(key), RemoteResourceWrapper(relationships[key]['data'], included))

                if 'links' in relationships[key]:
                    setattr(getattr(self, underscore(key)), 'links',
                            RemoteResourceWrapper(relationships[key]['links'], None))


class RemoteResourceListWrapper(list):

    def __init__(self, seq, included=None):
        super(RemoteResourceListWrapper, self).__init__()
        self.data = seq
        self.add_items_from_data(included)

    def add_items_from_data(self, included):
        map(lambda x: self.append(RemoteResourceWrapper(x, included)), self.data)


def get_route_from_fk(resource_type, pk=None):
    """Gets a fully qualified URL for a given resource_type, pk"""
    routes = requests.get(zc_settings.GATEWAY_ROOT_PATH).json()

    for route in routes.iterkeys():
        if 'resource_type' in routes[route] and routes[route]['resource_type'] == resource_type:
            if isinstance(pk, (list, set)):
                expanded = '{}?filter[id__in]={}'.format(expand(route, {}), ','.join([str(x) for x in pk]))
            else:
                expanded = expand(route, {'id': pk})
            return '{0}{1}'.format(routes[route]['domain'], expanded)

    raise RouteNotFoundException('No route for resource_type: "{0}"'.format(resource_type))


def make_service_request(service_name, endpoint, method=GET, data=None):
    """
    Makes a JWT token-authenticated service request to the URL provided.

    Args:
        service_name: name of the service making this request. e.g. mp-users
        endpoint: the url to use
        method: HTTP method. supported methods are defined at this module's global variables
        data: request payload in case we are making a POST request

    Returns: text content of the response
    """

    jwt_token = jwt_encode_handler(service_jwt_payload_handler(service_name))
    headers = {'Authorization': 'JWT {}'.format(jwt_token), 'Content-Type': 'application/vnd.api+json'}

    if method not in [GET, POST, PUT, PATCH]:
        raise UnsupportedHTTPMethodException(
            "Method {0} is not supported. service_name: {1}, endpoint: {2}".format(method, service_name, endpoint))

    response = getattr(requests, method)(endpoint, headers=headers, json=data)

    if 400 <= response.status_code < 600:
        http_error_msg = '{0} Error: {1} for {2}. Content: {3}'.format(
            response.status_code, response.reason, response.url, response.text)
        raise ServiceRequestException(http_error_msg)

    return response


def wrap_resource_from_response(response):
    json_response = ujson.loads(response.text)

    if 'data' not in json_response:
        msg = 'Error retrieving resource. Url: {0}. Content: {1}'.format(response.request.url, response.content)
        raise RemoteResourceException(msg)

    resource_data = json_response['data']
    included_raw = json_response.get('included')
    included_data = _included_to_dict(included_raw)
    if isinstance(resource_data, list):
        return RemoteResourceListWrapper(resource_data, included_data)
    return RemoteResourceWrapper(resource_data, included_data)


def _included_to_dict(included):
    data = {}

    if not included:
        return data

    for item in included:
        data[(item['type'], item['id'])] = item

    return data


def get_remote_resource(service_name, resource_type, pk, include=None, page_size=None):
    """A shortcut function to make a GET request to a remote service."""
    url = get_route_from_fk(resource_type, pk)

    params = {}
    if include:
        params['include'] = include

    if page_size:
        params['page_size'] = page_size

    if params:
        url = '{}?{}'.format(url, urllib.urlencode(params))

    response = make_service_request(service_name, url)
    wrapped_resource = wrap_resource_from_response(response)
    return wrapped_resource


def get_remote_resource_from_url(service_name, url):
    response = make_service_request(service_name, url)
    wrapped_resource = wrap_resource_from_response(response)
    return wrapped_resource



import re
from importlib import import_module

from mock import Mock

from django.contrib.admindocs.views import extract_views_from_urlpatterns, simplify_regex
from django.conf import settings

from .authentication import User
from .utils import jwt_payload_handler, service_jwt_payload_handler, jwt_encode_handler


def get_service_endpoint_urls(urlconfig=None, default_value='1'):
    """
    This function finds all endpoint urls in a service.

    Args:
        urlconfig: A django url config module to use. Defaults to settings.ROOT_URLCONF
        default_value: A string value to replace all url parameters with

    Returns:
        A list of urls with path parameters (i.e. <pk>) replaced with default_value
    """
    if not urlconfig:
        urlconfig = getattr(settings, 'ROOT_URLCONF')

    try:
        urlconfig_mod = import_module(urlconfig)
    except Exception as ex:
        raise Exception(
            "Unable to import url config module. Url Config: {0}. Message: {1}".format(urlconfig, ex.message))

    extracted_views = extract_views_from_urlpatterns(urlconfig_mod.urlpatterns)
    views_regex_url_patterns = [item[1] for item in extracted_views]
    simplified_regex_url_patterns = [simplify_regex(pattern) for pattern in views_regex_url_patterns]

    # Strip out urls we don't need to test.
    result_urls = []
    pattern = re.compile(r'<\w+>')

    for url in simplified_regex_url_patterns:
        if url.find('<format>') != -1:
            continue

        if url == u'/':
            continue

        if url == u'/health':
            continue

        parameters = pattern.findall(url)
        for param in parameters:
            url = url.replace(param, default_value)

        result_urls.append(url)

    return result_urls


class AuthenticationMixin:
    def create_user(self, roles, user_id):
        return User(roles=roles, pk=user_id)

    def create_user_token(self, user):
        payload = jwt_payload_handler(user)
        token = "JWT {}".format(jwt_encode_handler(payload))
        return token

    def create_service_token(self, service_name):
        payload = service_jwt_payload_handler(service_name)
        token = "JWT {}".format(jwt_encode_handler(payload))
        return token

    def get_staff_token(self, user_id):
        staff_user = self.create_user(['user', 'staff'], user_id)
        return self.create_user_token(staff_user)

    def get_user_token(self, user_id):
        user = self.create_user(['user'], user_id)
        return self.create_user_token(user)

    def get_guest_token(self, user_id):
        user = self.create_user(['user'], user_id)
        return self.create_user_token(user)

    def get_anonymous_token(self):
        return "JWT {}".format(jwt_encode_handler({'roles': ['anonymous']}))

    def get_service_token(self, service_name):
        return self.create_service_token(service_name)

    def authorize_as(self, user_type, **kwargs):
        method = getattr(self, 'get_%s_token' % user_type)
        if user_type == 'service':
            token = method(kwargs.get('service_name', 'Test'))
        elif user_type == 'anonymous':
            token = method()
        else:
            token = method(kwargs.get('user_id', 1))
        self.client.credentials(HTTP_AUTHORIZATION=token)


class PermissionTestMixin(object):
    def setUp(self):
        self.user = type('User', (Mock,), {'roles': [], 'id': 1})
        self.request = type('Request', (Mock,), {'user': self.user})
        self.view = type('View', (Mock,), {})
        self.obj = None

        # Create an instance of the Permission class in child class
        self.permission_obj = None

    def assert_has_permission(self, expected):
        has_perm = self.permission_obj.has_permission(self.request, self.view)
        self.assertEqual(has_perm, expected)

    def assert_has_object_permission(self, expected):
        has_perm = self.permission_obj.has_object_permission(self.request, self.view, self.obj)
        self.assertEqual(has_perm, expected)

    def assert_permission(self, expected):
        has_perm = (self.permission_obj.has_permission(self.request, self.view) and
                    self.permission_obj.has_object_permission(self.request, self.view, self.obj))
        self.assertEqual(has_perm, expected)



import jwt
from django.utils.encoding import smart_text
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework_jwt.settings import api_settings
from rest_framework_jwt.utils import jwt_decode_handler


class User(object):
    """
    A class that emulates Django's auth User, for use with microservices where
    the actual User is unavailable. Surfaces via `request.user`.
    """

    def __init__(self, **kwargs):
        self.roles = []
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])

    def is_authenticated(self):
        # Roles (i.e. anonymous, user, etc) are handled by permissions classes
        return True

    def get_roles(self):
        """
        For testing purposes only. Emulates `get_roles` in
        https://github.com/ZeroCater/mp-users/blob/master/users/models.py
        """
        return self.roles


class JWTAuthentication(BaseAuthentication):
    """
    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string specified in the setting
    `JWT_AUTH_HEADER_PREFIX`. For example:

        Authorization: JWT eyJhbGciOiAiSFMyNTYiLCAidHlwIj
    """
    www_authenticate_realm = 'api'

    @staticmethod
    def get_jwt_value(request):
        auth = get_authorization_header(request).split()
        auth_header_prefix = api_settings.JWT_AUTH_HEADER_PREFIX.lower()

        if not auth or smart_text(auth[0].lower()) != auth_header_prefix:
            return None

        if len(auth) == 1:  # pragma: no cover
            msg = 'Invalid Authorization header. No credentials provided.'
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:  # pragma: no cover
            msg = 'Invalid Authorization header. Credentials string should not contain spaces.'
            raise exceptions.AuthenticationFailed(msg)

        return auth[1]

    def authenticate(self, request):
        """
        Returns a two-tuple of `User` and token if a valid signature has been
        supplied using JWT-based authentication.  Otherwise returns `None`.
        """
        jwt_value = self.get_jwt_value(request)
        if jwt_value is None:
            raise exceptions.NotAuthenticated()

        try:
            payload = jwt_decode_handler(jwt_value)
        except jwt.ExpiredSignature:  # pragma: no cover
            msg = 'Signature has expired.'
            raise exceptions.AuthenticationFailed(msg)
        except jwt.DecodeError:  # pragma: no cover
            msg = 'Error decoding signature.'
            raise exceptions.AuthenticationFailed(msg)
        except jwt.InvalidTokenError:  # pragma: no cover
            raise exceptions.AuthenticationFailed()
        except Exception as ex:
            raise exceptions.AuthenticationFailed(ex.message)

        user = User(**payload)

        return user, jwt_value

    def authenticate_header(self, request):
        """
        Return a string to be used as the value of the `WWW-Authenticate`
        header in a `401 Unauthenticated` response, or `None` if the
        authentication scheme should return `403 Permission Denied` responses.
        """
        return '{0} realm="{1}"'.format(api_settings.JWT_AUTH_HEADER_PREFIX, self.www_authenticate_realm)



from collections import OrderedDict
import ujson

import six
from django.db.models.manager import BaseManager
from rest_framework_json_api.relations import ResourceRelatedField

from zc_common.remote_resource.models import RemoteResource


class RemoteResourceField(ResourceRelatedField):

    def __init__(self, related_resource_path=None, **kwargs):
        if 'model' not in kwargs:
            kwargs['model'] = RemoteResource
        if not kwargs.get('read_only', None):
            # The queryset is required to be not None, but not used
            #   due to the overriding of the methods below.
            kwargs['queryset'] = {}

        if related_resource_path is None:
            raise NameError('related_resource_path parameter must be provided')

        self.related_resource_path = related_resource_path

        super(RemoteResourceField, self).__init__(**kwargs)

    def get_links(self, obj=None, lookup_field='pk'):
        request = self.context.get('request', None)
        view = self.context.get('view', None)
        return_data = OrderedDict()

        kwargs = {lookup_field: getattr(obj, lookup_field) if obj else view.kwargs[lookup_field]}

        self_kwargs = kwargs.copy()
        self_kwargs.update({'related_field': self.field_name if self.field_name else self.parent.field_name})
        self_link = self.get_url('self', self.self_link_view_name, self_kwargs, request)

        # Construct the related link using the passed related_resource_path
        # self.source is the field name; getattr(obj, self.source) returns the
        # RemoteResource object or RelatedManager in the case of a to-many relationship.
        related_obj = getattr(obj, self.source)
        if related_obj and related_obj.id:
            if isinstance(related_obj, BaseManager):
                list_of_ids = related_obj.values_list('pk', flat=True)
                query_parameters = 'filter[id__in]={}'.format(','.join([str(pk) for pk in list_of_ids]))
                related_path = self.related_resource_path.format(pk=query_parameters)
            else:
                related_path = self.related_resource_path.format(pk=related_obj.id)
            related_link = request.build_absolute_uri(related_path)
        else:
            related_link = None

        if self_link:
            return_data.update({'self': self_link})
        if related_link:
            return_data.update({'related': related_link})
        return return_data

    def to_internal_value(self, data):
        if isinstance(data, six.text_type):
            try:
                data = ujson.loads(data)
            except ValueError:
                self.fail('incorrect_type', data_type=type(data).__name__)
        if not isinstance(data, dict):
            self.fail('incorrect_type', data_type=type(data).__name__)

        if 'type' not in data:
            self.fail('missing_type')

        if 'id' not in data:
            self.fail('missing_id')

        return RemoteResource(data['type'], data['id'])

    def to_representation(self, value):
        return OrderedDict([('type', value.type), ('id', str(value.id))])



import re
from django.conf import settings


class TaskRouter(object):
    def route_for_task(self, task, args=None, kwargs=None):
        if re.match('^microservice.*', task) is not None:
            return {'exchange': 'microservice-events',
                    'exchange_type': 'fanout',
                    'routing_key': ''}
        else:
            return {'exchange': 'default',
                    'exchange_type': 'direct',
                    'routing_key': settings.DEFAULT_QUEUE_NAME}



import ujson

import datetime
from inflection import camelize, underscore, pluralize
from rest_framework.test import APITestCase

from zc_common.jwt_auth.authentication import User
from zc_common.jwt_auth.utils import jwt_payload_handler, jwt_encode_handler

STAFF_PERMISSION_NAME = 'staff'
USER_PERMISSION_NAME = 'user'
SERVICE_PERMISSION_NAME = 'service'

USER = [USER_PERMISSION_NAME]
STAFF = [USER_PERMISSION_NAME, STAFF_PERMISSION_NAME]
SERVICE = [USER_PERMISSION_NAME, SERVICE_PERMISSION_NAME]


class ResponseTestCase(APITestCase):
    SUCCESS_HIGH_LEVEL_KEYS = ['data']
    SUCCESS_DATA_KEYS = ['id', 'type']

    FAILURE_HIGH_LEVEL_KEYS = ['errors']
    FAILURE_DATA_KEYS = ['status', 'source', 'detail']

    INSTANCE_TOP_LEVEL_KEYS = ['type', 'id', 'attributes']

    USER_ROLE = USER

    @staticmethod
    def convert_to_list(data):
        if isinstance(data, list):
            return data

        if isinstance(data, dict):
            return [data]

        try:
            return data.all()
        except AttributeError:
            return [data]

    def validate_instance_in_data(self, data, instance, attributes, relationship_keys=None):
        self.assertTrue(
            all(key in data for key in self.INSTANCE_TOP_LEVEL_KEYS))

        instance_attributes = data['attributes']
        for key in attributes:
            instance_attribute = getattr(instance, key)

            if hasattr(instance_attribute, '__call__'):
                instance_attribute = instance_attribute()

            if isinstance(instance_attribute, datetime.datetime):
                value = instance_attribute.isoformat()
                if value.endswith('+00:00'):
                    value = value[:-6] + 'Z'
                self.assertEqual(instance_attributes[camelize(key, False)], value)

            elif isinstance(instance_attribute, datetime.date):
                value = instance_attribute.isoformat()
                self.assertEqual(instance_attributes[camelize(key, False)], value)

            elif isinstance(instance_attribute, datetime.time):
                value = instance_attribute.isoformat()
                self.assertEqual(instance_attributes[camelize(key, False)], value)

            else:
                self.assertEqual(
                    instance_attributes[camelize(key, False)], instance_attribute)

        if not relationship_keys:
            return True

        relationships = data['relationships']

        for relationship_name in relationship_keys:
            name = camelize(relationship_name, False)
            self.assertTrue(name in relationships)
            data = self.convert_to_list(relationships[name]['data'])
            instance_relationship = self.convert_to_list(
                getattr(instance, relationship_name))

            for data_object, instance_object in zip(data, instance_relationship):
                self.assertEqual(data_object['id'], str(instance_object.id))

                try:
                    instance_type = instance_object.type
                except AttributeError:
                    instance_type = instance_object.__class__.__name__

                self.assertEqual(data_object['type'], instance_type)

    def validate_instance_in_response(self, response, instance, attributes, relationship_keys=None):
        data = self.load_json(response)['data']
        self.validate_instance_in_data(
            data, instance, attributes, relationship_keys)

    def validate_instance_list_in_response(self, response, instances, attributes, relationship_keys=None):
        instance_data = self.load_json(response)['data']
        self.assertEqual(len(instance_data), len(instances))

        instance_data.sort(key=lambda x: x['id'])
        list(instances).sort(key=lambda x: x.id)

        for data, instance in zip(instance_data, instances):
            self.validate_instance_in_data(
                data, instance, attributes, relationship_keys)

    def success_response_structure_test(self, response, status, relationship_keys=None):
        """
        This can be extended in the future to cover stricter validation of the
        response as follows:

        * Top level response MUST contain AT LEAST ONE of ['data', 'meta']
        * Top level response MUST contain ['links']  # Our requirement
            - Top level links MUST contain ['self']; MAY contain ['related']

        * Resource object MUST contain ['id', 'type']
        * Resource object MUST contain AT LEAST ONE of ['attributes', 'relationships']
        * Resource object MAY contain ['links', 'meta']

        * Relationship object MUST contain AT LEAST ONE of ['links', 'data', 'meta']
            - Relationship links object MUST contain AT LEAST ONE of ['self', 'related']
        """
        self.assertEqual(response.status_code, status)

        response_content = self.load_json(response)

        self.assertTrue(
            all(key in response_content for key in self.SUCCESS_HIGH_LEVEL_KEYS))

        for data in self.convert_to_list(response_content['data']):
            self.assertTrue(all(key in data for key in self.SUCCESS_DATA_KEYS))

            if relationship_keys:
                self.assertTrue('relationships' in data)
                relationships = data['relationships']
                self.assertTrue(
                    all(camelize(key, False) in relationships for key in relationship_keys))

                for relationship_name, relationship in relationships.iteritems():
                    self.assertTrue(
                        all(key in relationship for key in ['data', 'links']))

                    for relationship_data in self.convert_to_list(relationship['data']):
                        self.assertTrue(
                            all(key in relationship_data for key in ['type', 'id']))

                    links = relationship['links']

                    resource_pk = self.resource.pk if hasattr(self, 'resource') else '[0-9A-Za-z]*'

                    self.assertRegexpMatches(links['self'], r'^https?://.*/{}/{}/relationships/{}'.format(
                        self.resource_name, resource_pk, underscore(relationship_name)))

                    if hasattr(self, 'remote_relationship_keys') \
                            and relationship_name in self.remote_relationship_keys:
                        self.assertRegexpMatches(links['related'], r'^https?://.*/{}/\w'.format(
                            pluralize(underscore(self.get_remote_relationship_name(relationship_name)))))
                    else:
                        self.assertRegexpMatches(links['related'], r'^https?://.*/{}/{}/{}'.format(
                            self.resource_name, resource_pk, underscore(relationship_name)))

    def get_remote_relationship_name(self, relationship_name):
        """
        If there are any remote relationships whose related link doesn't map
        directly to its field name, a mapping dictionary can be declared on
        the test class named self.relationship_name_mapping.

        Example:
        self.relationship_name_mapping = {
            'homeAddress': 'address',
            'workAddress': 'address'
        }
        """
        if hasattr(self, 'relationship_name_mapping') and self.relationship_name_mapping.get(relationship_name):
            return self.relationship_name_mapping[relationship_name]
        return relationship_name

    def failure_response_structure_test(self, response, status):
        self.assertEqual(response.status_code, status)

        response_content = self.load_json(response)

        self.assertTrue(
            all(key in response_content for key in self.FAILURE_HIGH_LEVEL_KEYS))

        for error in response_content['errors']:
            self.assertTrue(
                all(key in error for key in self.FAILURE_DATA_KEYS))

    @staticmethod
    def load_json(response):
        return ujson.loads(response.content.decode())

    @staticmethod
    def generate_token(user):
        payload = jwt_payload_handler(user)
        token = jwt_encode_handler(payload)

        return "JWT {}".format(token)

    @staticmethod
    def generate_user(roles, user_id=None):
        if not user_id:
            user_id = '1'
        return User(pk=user_id, roles=roles)

    def client_request_auth(self, method, url, *args, **kwargs):
        roles = kwargs.pop('user_role', None)
        user_id = kwargs.pop('user_id', None)

        if not roles:
            return getattr(self.client, method)(url, *args, **kwargs)

        user = self.generate_user(roles, user_id)

        kwargs['HTTP_AUTHORIZATION'] = self.generate_token(user)

        return getattr(self.client, method)(url, *args, **kwargs)

    def client_get_auth(self, url, *args, **kwargs):
        return self.client_request_auth('get', url, *args, **kwargs)

    def client_post_auth(self, url, *args, **kwargs):
        return self.client_request_auth('post', url, *args, **kwargs)

    def client_patch_auth(self, url, *args, **kwargs):
        return self.client_request_auth('patch', url, *args, **kwargs)

    def client_delete_auth(self, url, *args, **kwargs):
        return self.client_request_auth('delete', url, *args, **kwargs)
